
#include "basic_solver.h"

// #define TRACE_PRINT_LOTS

void build_codemand_graph
(const index_set_vec& o, weighted_graph& g)
{
  assert(o.length() > 0);
  g.init(o.length());
  for (index_type i = 0; i < o.length(); i++)
    for (index_type j = i + 1; j < o.length(); j++) {
      index_type n = o[i].count_common(o[j]);
      if (n > 0) {
	g.add_undirected_edge(i, j, n);
      }
    }
}

void build_weighted_codemand_graph
(const index_set_vec& o, const index_matrix& c, bool maxc, weighted_graph& g)
{
  assert(o.length() > 0);
  g.init(o.length());
  for (index_type i = 0; i < o.length(); i++) {
    for (index_type j = i + 1; j < o.length(); j++) {
      index_type ew = 0;
      for (index_type k = 0; k < o[i].length(); k++)
	if (o[j].contains(o[i][k])) {
	  index_type w =
	    (c[i][o[i][k]] < c[j][o[i][k]] ? c[i][o[i][k]] : c[j][o[i][k]]);
	  if (maxc)
	    ew = (w > ew ? w : ew);
	  else
	    ew += w;
	}
      if (ew > 0) {
	g.add_undirected_edge(i, j, ew);
      }
    }

    index_type nw = 0;
    for (index_type k = 0; k < o[i].length(); k++) {
      index_type w = c[i][o[i][k]];
      if (maxc)
	nw = (w > nw ? w : nw);
      else
	nw += w;
    }
    g.set_weight(i, nw);
  }
}

void remove_common_requests
(index_type i, index_set& oi,
 index_type j, index_set& oj,
 const index_matrix& c)
{
  index_set ri;
  index_set rj;
  for (index_type k = 0; k < oi.length(); k++) if (oj.contains(oi[k])) {
    if (c[i][oi[k]] < c[j][oi[k]])
      ri.insert(oi[k]);
    else
      rj.insert(oi[k]);
  }
  oi.subtract(ri);
  oj.subtract(rj);
}

void write_codemand_graph(std::ostream& s, const index_set_vec& o)
{
  s << "graph \"Co-Demand Graph\" {" << std::endl;
  s << "overlap=false;" << std::endl;
  s << "edge [len=1.0];" << std::endl;
  s << "node [shape=circle];" << std::endl;

  for (index_type i = 0; i < o.length(); i++)
    for (index_type j = i + 1; j < o.length(); j++) {
      index_set p(o[i]);
      p.intersect(o[j]);
      if (!p.empty()) {
	s << i << " -- " << j << " [label=\"" << p << "\"];" << std::endl;
      }
    }
  s << "}" << std::endl;
}

void order_first_and_last
(const index_set& o, const index_vec& is, index_type& f, index_type& l)
{
  f = no_such_index;
  l = no_such_index;
  for (index_type i = 0; i < o.length(); i++) {
    if (f == no_such_index) 
      f = is[o[i]];
    else if (is[o[i]] > f)
      f = is[o[i]];
    if (l == no_such_index) 
      l = is[o[i]];
    else if (is[o[i]] < l)
      l = is[o[i]];
  }
}

void preprocess
(index_type p,
 const index_set_vec& o,
 index_type& rp,
 index_set_vec& ro,
 index_set_vec& pm)
{
  index_set_vec ofp(EMPTYSET, p);
  for (index_type k = 0; k < o.length(); k++)
    for (index_type i = 0; i < o[k].length(); i++)
      ofp[o[k][i]].insert(k);
  graph dominance(p);
  for (index_type k = 0; k < p; k++) {
    for (index_type i = 0; i < p; i++)
      if ((i != k) && ofp[i].contains(ofp[k]) &&
	  (!ofp[k].contains(ofp[i]) || (k < i))) {
	dominance.add_edge(i, k);
      }
  }
  std::cerr << "product dominance graph: " << dominance << std::endl;
  index_set rps;
  for (index_type k = 0; k < p; k++)
    if (dominance.in_degree(k) == 0)
      rps.insert(k);
  std::cerr << "reduced product set: " << rps << std::endl;
  index_vec p_map;
  mapping::invert_map(rps, p_map, p);
  std::cerr << "inverse map: " << p_map << std::endl;
  ro.assign_value(EMPTYSET, o.length());
  for (index_type k = 0; k < o.length(); k++) {
    for (index_type i = 0; i < o[k].length(); i++)
      if (p_map[o[k][i]] != no_such_index)
	ro[k].insert(p_map[o[k][i]]);
    std::cerr << "reduced order #" << k << ": " << ro[k] << std::endl;
  }
  rp = rps.length();
  std::cerr << "#products reduced to " << rp
	    << " by preprocessing" << std::endl;
  pm.assign_value(EMPTYSET, rp);
  for (index_type k = 0; k < rp; k++) {
    dominance.descendants(rps[k], pm[k]);
    pm[k].insert(rps[k]);
  }
}

void postprocess
(const index_vec& s_in,
 const index_set_vec& pm,
 index_type np,
 index_vec& s_out)
{
  s_out.set_length(0);
  bool_vec in_seq(false, np);
  for (index_type k = 0; k < s_in.length(); k++) {
    assert(s_in[k] < pm.length());
    for (index_type i = 0; i < pm[k].length(); i++)
      if (!in_seq[pm[k][i]]) {
	s_out.append(pm[k][i]);
	in_seq[pm[k][i]] = true;
      }
  }
}

#define LARGE_PRIME 2147483629

index_type set_compare::length = 0;

void set_compare::init(index_type l)
{
  length = l;
}

bool set_compare::operator()(const bool_vec& s0, const bool_vec& s1) const
{
  for (index_type k = 0; k < length; k++) {
    if (s0[k] != s1[k]) return false;
  }
  return true;
}

index_type set_hash::length = 0;
index_vec  set_hash::factors(0, 0);

void set_hash::init(index_type l)
{
  length = l;
  factors.assign_value(2, length);
  for (index_type k = 1; k < length; k++) {
    factors[k] = ((2 * factors[k - 1]) % LARGE_PRIME);
  }
}

index_type set_hash::operator()(const bool_vec& s0) const
{
  index_type v = 0;
  for (index_type k = 0; k < length; k++) if (s0[k]) {
    v = ((v + factors[k]) % LARGE_PRIME);
  }
  return v;
}

BasicSolver::BasicSolver
(index_type p, const index_set_vec& o)
  : n_products(p), orders(o), n_solutions(0)
{
  assert(n_products > 0);
  assert(orders.length() > 0);
  set_compare::init(n_products);
  set_hash::init(n_products);
}

BasicSolver::~BasicSolver()
{
  // done
}

index_type BasicSolver::active_count(index_type p, bool_vec& l)
{
  index_type n = 0;
  index_set  a;
  for (index_type k = 0; k < orders.length(); k++) {
    bool is_a = false;
    for (index_type i = 0; (i < orders[k].length()) && !is_a; i++) {
      if (orders[k][i] == p) {
	a.insert(k);
	n += 1;
	is_a = true;
      }
      for (index_type j = i + 1; (j < orders[k].length()) && !is_a; j++)
	if ((l[orders[k][i]] && !l[orders[k][j]]) ||
	    (!l[orders[k][i]] && l[orders[k][j]])) {
	  a.insert(k);
	  n += 1;
	  is_a = true;
	}
    }
  }
  assert(n == a.length());
  return n;
}

index_type BasicSolver::subset_min_active_count
(bool_vec& s, bool_vec& x, index_type& v_min)
{
  index_type p_min = no_such_index;
  v_min = orders.length() + 1;
  for (index_type k = 0; k < n_products; k++) if (s[k] && !x[k]) {
    s[k] = false;
    index_type v = active_count(k, s);
    s[k] = true;
    if (v < v_min) {
      p_min = k;
      v_min = v;
    }
  }
  assert(p_min < n_products);
  assert(s[p_min]);
  return p_min;
}

void BasicSolver::print_state(std::ostream& t, bool_vec& s, index_type n)
{
  t << "[" << n << ":";
  for (index_type k = 0; k < n_products; k++)
    if (s[k])
      t << " " << k;
  t << "]";
}

index_type BasicSolver::stacks
(bool_vec& s, index_type n, index_type L, index_type U, bool all)
{
  if (n == 0) {
    n_solutions += 1;
    return 0;
  }
  set_map::iterator v = s_table.find(s);
  if (v != s_table.end()) {
    return v->second;
  }
  index_type min = U + 1;
  bool_vec tried(false, n_products);
  while (((min > L) || all) && (tried.count(true) < n)) {
#ifdef TRACE_PRINT_LOTS
    print_state(std::cerr, s, n);
    std::cerr << ": min = " << min << ", L = " << L << ", U = " << U
	      << ", tried = " << tried.count(true) << std::endl;
#endif
    index_type c_min;
    index_type p_min = subset_min_active_count(s, tried, c_min);
    tried[p_min] = true;
    if ((c_min < min) || (all && (c_min == min))) {
      s[p_min] = false;
      index_type rp = stacks(s, n - 1, L, U, all);
      index_type sp = (rp >= c_min ? rp : c_min);
      if (sp < min) min = sp;
      if (all && (min < L)) min = L;
      s[p_min] = true;
    }
  }
  s_table[s] = min;
  return min;
}

index_type BasicSolver::stacks(index_type L, index_type U, bool all)
{
  bool_vec s(true, n_products);
  return stacks(s, n_products, L, U, all);
}

index_type BasicSolver::morestacks
(bool_vec& s, index_type n, index_type L, index_type U, index_type n_min)
{
  if (n == 0) {
    n_solutions += 1;
    return 0;
  }
  set_map::iterator v = s_table.find(s);
  if (v != s_table.end()) {
    if (v->second > L)
      return v->second;
  }
  index_type min = U + 1;
  bool_vec tried(false, n_products);
  while ((n_solutions <= n_min) && (tried.count(true) < n)) {
#ifdef TRACE_PRINT_LOTS
    print_state(std::cerr, s, n);
    std::cerr << ": min = " << min << ", L = " << L << ", U = " << U
	      << ", tried = " << tried.count(true) << std::endl;
#endif
    index_type c_min;
    index_type p_min = subset_min_active_count(s, tried, c_min);
    tried[p_min] = true;
    if (c_min < min) {
      s[p_min] = false;
      index_type rp = morestacks(s, n - 1, L, U, n_min);
      index_type sp = (rp >= c_min ? rp : c_min);
      if (sp < min) min = sp;
      s[p_min] = true;
    }
  }
  s_table[s] = min;
  return min;
}

index_type BasicSolver::morestacks(index_type L, index_type U, index_type n)
{
  bool_vec s(true, n_products);
  index_type n_target = n_solutions + n;
  n_solutions = 0;
  return morestacks(s, n_products, L, U, n_target);
}

void BasicSolver::clear()
{
  s_table.clear();
  n_solutions = 0;
}

bool BasicSolver::extract
(bool_vec& s, index_type m, index_vec& seq)
{
  index_type n = s.count(true);
  if (n == 1) {
    for (index_type k = 0; k < n_products; k++) if (s[k]) {
      seq[n_products - 1] = k;
      return true;
    }
  }
  for (index_type k = 0; k < n_products; k++) if (s[k]) {
    s[k] = false;
    set_map::iterator v = s_table.find(s);
    if (v != s_table.end()) {
#ifdef TRACE_PRINT_LOTS
      std::cerr << "|s| = " << n
		<< ", k = " << k
		<< ", v = " << v->second
		<< ", max = " << m
		<< std::endl;
#endif
      if (v->second <= m) {
	bool ok = extract(s, m, seq);
	if (ok) {
	  seq[n_products - n] = k;
	  s[k] = true;
	  return true;
	}
      }
    }
    s[k] = true;
  }
  return false;
}

bool BasicSolver::extract_next
(bool_vec& s, index_type m, index_vec& seq)
{
  index_type n = s.count(true);
  if (n == 1) {
    return false;
  }
  index_type c = seq[n_products - n];
  s[c] = false;
  bool ok = extract_next(s, m, seq);
  if (ok) return true;
  s[c] = true;
  for (index_type k = c + 1; k < n_products; k++) if (s[k]) {
    s[k] = false;
    set_map::iterator v = s_table.find(s);
    if (v != s_table.end()) {
#ifdef TRACE_PRINT_LOTS
      std::cerr << "|s| = " << n
		<< ", k = " << k
		<< ", v = " << v->second
		<< ", max = " << m
		<< std::endl;
#endif
      if (v->second <= m) {
	bool ok = extract(s, m, seq);
	if (ok) {
	  seq[n_products - n] = k;
	  s[k] = true;
	  return true;
	}
      }
    }
    s[k] = true;
  }
  return false;
}

bool BasicSolver::extract(index_type m, index_vec& seq)
{
  seq.assign_value(no_such_index, n_products);
  bool_vec s(true, n_products);
  return extract(s, m, seq);
}

bool BasicSolver::extract_next(index_type m, index_vec& seq)
{
  bool_vec s(true, n_products);
  return extract_next(s, m, seq);
}
