
#include "gecode/kernel.hh"
#include "gecode/int.hh"
#include "gecode/set.hh"
#include "gecode/minimodel.hh"
#include "gecode/search.hh"

#include "problem_io.h"
#include "basic_solver.h"

#define VERBOSE
#define PRE_ASSIGN
// #define RELAXED
#define END_POINT_DURATION 0

class ColorCDG : public Gecode::Space
{
 protected:
  HSPS::index_type n_products;
  const HSPS::index_set_vec& orders;
  const HSPS::index_vec& duration;
  const HSPS::graph& cdg;
  const HSPS::index_vec& d_min;
  HSPS::index_type n_stacks;

  Gecode::IntVarArray color;
  Gecode::IntVarArray first;
#ifdef RELAXED
  Gecode::SetVarArray parts;
  Gecode::IntVarArray p_cost;
#else
  Gecode::IntVarArray po;
  Gecode::IntVarArray p_start;
  Gecode::IntVarArray p_finish;
#endif
  Gecode::IntVar makespan;

 public:
  ColorCDG(HSPS::index_type np,
	   const HSPS::index_set_vec& o,
	   const HSPS::index_vec& d1,
	   const HSPS::graph& g,
	   const HSPS::index_vec& d2,
	   HSPS::index_type t_max,
	   HSPS::index_type ns,
	   bool opt_pre_assign,
	   bool opt_apx_clique);
  ColorCDG(bool share, ColorCDG& s);

  virtual Gecode::Space* copy(bool share);
  int  solution_cost();
  void constrain(Gecode::Space* s);
  virtual void print();
};

ColorCDG::ColorCDG
(HSPS::index_type np,
 const HSPS::index_set_vec& o,
 const HSPS::index_vec& d1,
 const HSPS::graph& g,
 const HSPS::index_vec& d2,
 HSPS::index_type t_max,
 HSPS::index_type ns,
 bool opt_pre_assign,
 bool opt_apx_clique)
  : n_products(np), orders(o), duration(d1),
    cdg(g), d_min(d2), n_stacks(ns),
    color(this, cdg.size(), 0, n_stacks - 1),
    first(this, n_stacks, 0, cdg.size() - 1),
#ifdef RELAXED
    parts(this, n_stacks),
    p_cost(this, n_stacks, 0, t_max),
#else
    po(this, cdg.size(), 0, (cdg.size() / n_stacks)),
    p_start(this, n_products, 1, t_max),
    p_finish(this, n_products, 1, t_max),
#endif
    makespan(this, 0, t_max)
{
  assert(d_min.length() == cdg.size());

  // basic coloring constraints
  for (index_type i = 0; i < cdg.size(); i++)
    for (index_type j = i + 1; j < cdg.size(); j++)
      if (cdg.adjacent(i, j))
	post(this, (color[i] != color[j]));

#ifdef RELAXED
  // dual model & cost function
  channel(this, color, parts);
  Gecode::IntArgs e(cdg.size());
  Gecode::IntArgs w(cdg.size());
  for (HSPS::index_type i = 0; i < cdg.size(); i++) {
    e[i] = i;
    w[i] = (d_min[i] + (2 * END_POINT_DURATION));
  }
  for (HSPS::index_type j = 0; j < n_stacks; j++) {
    weights(this, e, w, parts[j], p_cost[j]);
    post(this, (makespan >= p_cost[j]));
  }
#else
  // duration constraints
  for (index_type i = 0; i < n_products; i++)
    post(this, (p_finish[i] == (p_start[i] + duration[i] - 1)));

  // partial ordering of orders implied by coloring
  for (index_type i = 0; i < cdg.size(); i++)
    for (index_type j = i + 1; j < cdg.size(); j++)
      post(this, Gecode::tt((color[i] != color[j]) || (po[i] != po[j])));

  // partial ordering of products implied by ordering of orders
  for (index_type oi = 0; oi < cdg.size(); oi++)
    for (index_type oj = 0; oj < cdg.size(); oj++)
      for (index_type pi = 0; pi < orders[oi].size(); pi++)
	for (index_type pj = 0; pj < orders[oj].size(); pj++)
	  post(this, Gecode::tt((po[oi] >= po[oj]) ||
				(color[oi] != color[oj]) ||
				(p_finish[orders[oi][pi]] <
				 p_start[orders[oj][pj]])));

  // cost function
  for (index_type i = 0; i < n_products; i++)
    post(this, (makespan >= p_finish[i]));
#endif

  // set up for symmetry breaking constraints
  for (index_type i = 0; i < cdg.size(); i++)
    for (index_type j = 0; j < n_stacks; j++)
      post(this, Gecode::tt((color[i] != j) || (first[j] <= i)));
  for (index_type i = 0; i < cdg.size(); i++)
    for (index_type j = 0; j < n_stacks; j++)
      post(this, Gecode::tt((first[j] != i) || (color[i] == j)));

  // pre-assignment
  if (opt_pre_assign) {
    HSPS::index_set q;
    std::cerr << "searching for clique..." << std::endl;
    if (opt_apx_clique) {
      graph ccdg(cdg);
      ccdg.complement();
      ccdg.apx_independent_set(q);
    }
    else {
      cdg.maximal_clique(q);
    }
    std::cerr << "found clique of size " << q.length() << std::endl;
    assert(q.length() <= n_stacks);
    for (HSPS::index_type i = 0; i < q.length(); i++)
      eq(this, color[q[i]], i);
    for (index_type j = q.length(); j < n_stacks - 1; j++)
      post(this, (first[j] < first[j + 1]));
  }
  else { // full symmetry breaking
    for (index_type j = 0; j < n_stacks - 1; j++)
      post(this, (first[j] < first[j + 1]));
  }

  branch(this, color, Gecode::BVAR_SIZE_MIN, Gecode::BVAL_MIN);
#ifndef RELAXED
  branch(this, po, Gecode::BVAR_SIZE_MIN, Gecode::BVAL_MIN);
#endif
}

ColorCDG::ColorCDG(bool share, ColorCDG& cc)
  : Gecode::Space(share, cc),
    n_products(cc.n_products),
    orders(cc.orders),
    duration(cc.duration),
    cdg(cc.cdg),
    d_min(cc.d_min),
    n_stacks(cc.n_stacks)
{
  color.update(this, share, cc.color);
  first.update(this, share, cc.first);
#ifdef RELAXED
  parts.update(this, share, cc.parts);
  p_cost.update(this, share, cc.p_cost);
#else
  po.update(this, share, cc.po);
  p_start.update(this, share, cc.p_start);
  p_finish.update(this, share, cc.p_finish);
#endif
  makespan.update(this, share, cc.makespan);
}

Gecode::Space* ColorCDG::copy(bool share)
{
  return new ColorCDG(share, *this);
}

int ColorCDG::solution_cost()
{
  return makespan.min();
}

void ColorCDG::constrain(Gecode::Space* s)
{
  int s_cost = ((ColorCDG*)s)->solution_cost();
  post(this, (makespan < s_cost));
  for (index_type i = 0; i < orders.length(); i++)
    for (index_type j = i + 1; j < orders.length(); j++)
      if ((d_min[i] + d_min[j] + (4 * END_POINT_DURATION)) > s_cost)
	post(this, (color[i] != color[j]));
}

void ColorCDG::print()
{
  std::cerr << "color = ";
  for (HSPS::index_type k = 0; k < n_stacks; k++) {
    if (k > 0) std::cerr << ", ";
    std::cerr << "{";
    bool f = true;
    for (HSPS::index_type i = 0; i < cdg.size(); i++)
      if (color[i].val() == k) {
	if (!f) std::cerr << ",";
	std::cerr << i;
	f = false;
      }
    std::cerr << "}";
  }
#ifndef RELAXED
  std::cerr << ", po = [";
  for (HSPS::index_type i = 0; i < cdg.size(); i++) {
    if (i > 0) std::cerr << ",";
    std::cerr << po[i].val();
  }
  std::cerr << "], schedule = ";
  for (HSPS::index_type i = 0; i < n_products; i++) {
    if (i > 0) std::cerr << ",";
    std::cerr << p_start[i].min() << "-" << p_finish[i].min();
    // std::cerr << "[" << p_start[i].min() << "," << p_start[i].max() << "]-["
    //	      << p_finish[i].min() << "," << p_finish[i].max() << "]";
  }
#endif
  std::cerr << ", cost = " << makespan.min() << std::endl;
}

int main(int argc, char** argv)
{
  HSPS::index_type    n_products;
  HSPS::index_set_vec orders;
  HSPS::index_vec     durations;
  HSPS::index_type    n_stacks = HSPS::no_such_index;
  HSPS::index_type    t_max = index_type_max;

  bool opt_pre_assign = true;
  bool opt_apx_clique = false;
  bool opt_unit_dur = false;

  for (HSPS::index_type k = 1; k < argc; k++) {
    if (strcmp(argv[k], "-no-pre") == 0) {
      opt_pre_assign = false;
    }
    else if (strcmp(argv[k], "-ac") == 0) {
      opt_apx_clique = true;
    }
    else if (strcmp(argv[k], "-no-dur") == 0) {
      opt_unit_dur = true;
    }
    else if ((strcmp(argv[k], "-tmax") == 0) && (k < argc - 1)) {
      t_max = atoi(argv[++k]);
    }
    else {
      n_stacks = atoi(argv[k]);
    }
  }

  if (n_stacks == HSPS::no_such_index) {
    std::cerr << argv[0] << " [-no-pre] #stacks" << std::endl;
    exit(0);
  }

  std::cerr << "reading problem..." << std::endl;
  read_basic_problem(n_products, orders);
  if (opt_unit_dur) {
    durations.assign_value(1, n_products);
  }
  else {
    read_durations(n_products, durations);
  }

  std::cerr << "problem read" << std::endl;
  std::cerr << "#products = " << n_products << std::endl;
  std::cerr << "durations = " << durations << std::endl;
  std::cerr << "#orders = " << orders.length() << std::endl;
  for (index_type k = 0; k < orders.length(); k++) {
    std::cerr << "order #" << k << ": " << orders[k] << std::endl;
  }

  weighted_graph cdg;
  build_codemand_graph(orders, cdg);
  HSPS::index_type t_sum = 0;
  for (index_type i = 0; i < n_products; i++) {
    t_sum += durations[i];
  }
  if (t_sum < t_max) t_max = t_sum;

  HSPS::index_vec d_min(0, orders.length());
  for (index_type k = 0; k < orders.length(); k++)
    for (index_type i = 0; i < orders[k].length(); i++)
      if (durations[orders[k][i]] > d_min[k])
	d_min[k] = durations[orders[k][i]];

  std::cerr << "creating model..." << std::endl;
  ColorCDG* cc = 0;
  HSPS::index_type n_sol = 0;
  int best = 0;
  try {
    cc = new ColorCDG(n_products, orders, durations,
		      cdg, d_min, t_max, n_stacks,
		      opt_pre_assign, opt_apx_clique);

    if (cc->status() == Gecode::SS_FAILED) {
      std::cerr << "status == FAILED (over before it started!)" << std::endl;
      exit(0);
    }

    std::cerr << "creating engine..." << std::endl;
    // Gecode::DFS<ColorCDG> engine(cc);
    Gecode::BAB<ColorCDG> engine(cc);

    std::cerr << "searching..." << std::endl;
    ColorCDG* sol = engine.next();
    while (sol != NULL) {
#ifdef VERBOSE
      std::cerr << "found a solution (status = " << sol->status() << "): ";
      sol->print();
      std::cerr << "stats: " << engine.statistics().propagate
		<< ", " << engine.statistics().fail
		<< ", " << engine.statistics().clone
		<< ", " << engine.statistics().commit
		<< ", " << engine.statistics().memory
		<< std::endl;
#endif
      if ((n_sol == 0) || (sol->solution_cost() < best))
	best = sol->solution_cost();
      delete sol;
      n_sol += 1;
      sol = engine.next();
    }
    std::cerr << "no more solutions" << std::endl;

    std::cerr << n_sol << " solutions" << std::endl;
    std::cerr << "min cost = " << best << std::endl;

    std::cerr << "propagation steps: "
	      << engine.statistics().propagate << std::endl;
    std::cerr << "failed nodes:      "
	      << engine.statistics().fail << std::endl;
    std::cerr << "clones:            "
	      << engine.statistics().clone << std::endl;
    std::cerr << "commit operations: "
	      << engine.statistics().commit << std::endl;
    std::cerr << "peak memory:       "
	      << engine.statistics().memory << std::endl;
  }
  catch (Gecode::Exception e) {
    std::cerr << "exception " << e.what() << " caught" << std::endl;
    exit(1);
  }
}
