
#include "basic_solver.h"
#include "soft_solver.h"
#include "problem_io.h"
#include "stats.h"

#include <fstream>

int main(int argc, char* argv[])
{
  index_type    n_products;
  index_set_vec orders;
  index_matrix  costs;
  index_vec     schedule;
  bool          test_schedule = false;
  bool          opt_plan = false;
  bool          opt_soft_exp = false;
  bool          opt_soft_lin = false;
  index_type    soft_stack_limit = 0;
  index_type    soft_max_loss = 0;

  bb_verbose = true;
  for (index_type k = 0; k < argc; k++) {
    if (strcmp(argv[k],"-p") == 0) {
      opt_plan = true;
    }
    else if (strcmp(argv[k],"-pp") == 0) {
      bb_print_plan = true;
    }
    else if (strcmp(argv[k],"-s") == 0) {
      test_schedule = true;
    }
    else if ((strcmp(argv[k],"-soft-exp") == 0) && (k < (argc - 2))) {
      opt_soft_exp = true;
      soft_stack_limit = atoi(argv[++k]);
      soft_max_loss = atoi(argv[++k]);
    }
    else if ((strcmp(argv[k],"-soft-lin") == 0) && (k < (argc - 2))) {
      opt_soft_lin = true;
      soft_stack_limit = atoi(argv[++k]);
      soft_max_loss = atoi(argv[++k]);
    }
  }

//   if (!opt_soft_exp && !opt_soft_lin) {
//     std::cerr << "bad user! (must specify -soft-exp <l> <b> or -soft-lin <l> <b>)" << std::endl;
//     exit(0);
//   }

  Stopwatch stats;

  read_weighted_problem(n_products, orders, costs);
  read_schedule(n_products, schedule);
  index_vec invs;
  mapping::invert_map(schedule, invs);

  std::cerr << "problem read" << std::endl;
  std::cerr << "#products = " << n_products << std::endl;
  std::cerr << "#orders = " << orders.length() << std::endl;
  for (index_type k = 0; k < orders.length(); k++) {
    std::cerr << "order #" << k << ": " << orders[k] << std::endl;
  }
  std::cerr << "schedule = " << schedule << std::endl;
  std::cerr << "inv. schedule: " << invs << std::endl;

  if (test_schedule) {
    index_vec f;
    index_vec l;
    index_vec n;
    compute_open_schedule(orders, schedule, f, l, n);
    std::cerr << "f = " << f << std::endl;
    std::cerr << "l = " << l << std::endl;
    std::cerr << "n = " << n << std::endl;
    write_open_schedule(std::cout, orders, schedule, &costs);
  }

  if (opt_soft_exp) {
    std::cerr << "computing min loss (exp.) with stack limit = "
	      << soft_stack_limit << " and bound = " << soft_max_loss
	      << "..." << std::endl;
    index_set_vec o_min(EMPTYSET, 0);
    stats.start();
    index_type v =
      min_exp_loss(orders, schedule, soft_stack_limit, soft_max_loss, o_min);
    stats.stop();
    std::cerr << "min loss = " << v
	      << " (" << stats.time() << " seconds)"
	      << std::endl;
    if (o_min.length() == 0) {
      std::cerr << "error: no solution within this stack limit and bound!"
		<< std::endl;
      exit(0);
    }
    assert(o_min.length() == orders.length());
    orders.assign_copy(o_min);
  }

  else if (opt_soft_lin) {
    std::cerr << "computing min loss (linear) with stack limit = "
	      << soft_stack_limit << " and bound = " << soft_max_loss
	      << "..." << std::endl;
    index_set_vec o_min(EMPTYSET, 0);
    stats.start();
    index_type v =
      min_lin_loss(orders, schedule, soft_stack_limit, soft_max_loss, o_min);
    stats.stop();
    std::cerr << "min loss = " << v
	      << " (" << bb_lin_calls << " calls, "
	      << stats.time() << " seconds, "
	      << (stats.time() / (double)bb_lin_calls) << " seconds/call)"
	      << std::endl;
    if (o_min.length() == 0) {
      std::cerr << "error: no solution within this stack limit and bound!"
		<< std::endl;
      exit(0);
    }
    assert(o_min.length() == orders.length());
    for (index_type k = 0; k < orders.length(); k++) {
      index_set o(orders[k]);
      o.subtract(o_min[k]);
      if (!o.empty()) {
	std::cerr << " products " << o << " removed from order #" << k
		  << std::endl;
      }
    }
    orders.assign_copy(o_min);
  }

  if (opt_plan) {
    if (opt_soft_exp || opt_soft_lin) {
      write_sp_plan(std::cout, orders, schedule, soft_stack_limit, false);
    }
    else {
      index_vec f;
      index_vec l;
      index_vec n;
      compute_open_schedule(orders, schedule, f, l, n);
      index_type n_max = n[n.arg_max()];
      write_seq_plan(std::cout, orders, schedule, n_max);
    }
  }
}
