
#include "gecode/kernel.hh"
#include "gecode/int.hh"
#include "gecode/minimodel.hh"
#include "gecode/search.hh"

#include "problem_io.h"

#define RELAXED
// #define EXTRA_CARD_CONSTRAINTS

// #define MY_SEARCH
#define STOP_AT_FIRST_FAIL

count_type n_backtracks = 0;

class OpenStacksTime : public Gecode::Space
{
 protected:
  index_type n_products;
  index_type n_orders;
  index_type n_stacks;
  index_type t_max;

  const index_set_vec& orders;
  const index_vec&     duration;

  Gecode::IntVarArray start_order;
  Gecode::IntVarArray finish_order;
  Gecode::IntVarArray open_order;
  Gecode::IntVarArray start_product;

  // extra variables needed to encode cardinality constraints
#ifdef EXTRA_CARD_CONSTRAINTS
  Gecode::IntVarArray finish_product;
  Gecode::BoolVarArray start_before_make;
  Gecode::BoolVarArray finish_after_make;
  Gecode::IntVarArray  n_before;
  Gecode::IntVarArray  n_after;
#endif

 public:
  OpenStacksTime(HSPS::index_type np,
		 const HSPS::index_set_vec& o,
		 const HSPS::index_vec& d,
		 HSPS::index_type ns,
		 HSPS::index_type tm);
  OpenStacksTime(bool share, OpenStacksTime& s);

  virtual Gecode::Space* copy(bool share);
  virtual void print();

  // constructive search
  void construct(const graph& cdg,
		 bool_vec& o_open,
		 bool_vec& o_done);
};

OpenStacksTime::OpenStacksTime
(HSPS::index_type np,
 const HSPS::index_set_vec& o,
 const HSPS::index_vec& d,
 HSPS::index_type ns,
 HSPS::index_type tm)
  : n_products(np), n_orders(o.length()), n_stacks(ns), t_max(tm),
    orders(o), duration(d),
    start_order(this, n_orders, 0, t_max - 1),
    finish_order(this, n_orders, 0, t_max - 1),
    open_order(this, n_orders, 0, t_max - 1),
    start_product(this, n_products, 0, t_max - 1)
#ifdef EXTRA_CARD_CONSTRAINTS
  , finish_product(this, n_products, 0, t_max - 1),
    start_before_make(this, n_orders * n_products, 0, 1),
    finish_after_make(this, n_orders * n_products, 0, 1),
    n_before(this, n_products, 0, n_orders),
    n_after(this, n_products, 0, n_orders)
#endif
{
  // basic constraints between orders and products
  Gecode::IntArgs use(n_orders);
  for (int i = 0; i < orders.length(); i++) {
    index_type d_min = 0;
    for (int j = 0; j < orders[i].length(); j++) {
      post(this, (start_order[i] <= start_product[orders[i][j]]),
	   Gecode::ICL_DOM);
      post(this, ((start_product[orders[i][j]] + d[orders[i][j]])
      		  <= finish_order[i]),
	   Gecode::ICL_DOM);
      if (d[orders[i][j]] > d_min) d_min = d[orders[i][j]];
    }
    post(this, (start_order[i] + open_order[i] == finish_order[i]),
	 Gecode::ICL_DOM);
    post(this, (open_order[i] >= d_min), Gecode::ICL_DOM);
    use[i] = 1;
  }

  // encode stack limit as a cumulative constraint over orders
  cumulative(this, start_order, open_order, use, n_stacks,
  	     true, Gecode::ICL_DOM);

#ifndef RELAXED
  // all order end-points must be distinct
  Gecode::IntVarArgs points(2 * n_orders);
  for (index_type i = 0; i < n_orders; i++) {
    points[i] = start_order[i];
    points[i + n_orders] = finish_order[i];
  }
  distinct(this, points, Gecode::ICL_DOM);

  // extra constraints on product start times
  for (index_type i = 0; i < n_products; i++) {
#ifdef EXTRA_CARD_CONSTRAINTS
    post(this, (finish_product[i] == (start_product[i] + d[i])),
	 Gecode::ICL_DOM);
#else
    index_type n_in = 0;
    for (int j = 0; j < orders.length(); j++)
      if (orders[j].contains(i))
	n_in += 1;
    post(this, (start_product[i] > n_in), Gecode::ICL_DOM);
    post(this, ((start_product[i] + d[i]) < (t_max - n_in)),Gecode::ICL_DOM);
#endif
  }
#endif

#ifdef EXTRA_CARD_CONSTRAINTS
  for (index_type i = 0; i < n_products; i++) {
    Gecode::IntVarArgs args(n_orders);
    for (index_type j = 0; j < n_orders; j++) {
      rel(this, start_order[j], Gecode::IRT_LE, start_product[i],
	  start_before_make[(i * n_orders) + j], Gecode::ICL_DOM);
      args[j] = start_before_make[(i * n_orders) + j];
    }
    count(this, args, 1, Gecode::IRT_EQ, n_before[i]);
    post(this, (start_product[i] > n_before[i]), Gecode::ICL_DOM);
  }

  for (index_type i = 0; i < n_products; i++) {
    Gecode::IntVarArgs args(n_orders);
    for (index_type j = 0; j < n_orders; j++) {
      rel(this, finish_order[j], Gecode::IRT_GR, finish_product[i],
	  finish_after_make[(i * n_orders) + j], Gecode::ICL_DOM);
      args[j] = finish_after_make[(i * n_orders) + j];
    }
    count(this, args, 1, Gecode::IRT_EQ, n_after[i]);
    post(this, (finish_product[i] <= (t_max - n_after[i])), Gecode::ICL_DOM);
  }
#endif

#ifndef MY_SEARCH
  // try BVAR_REGRET_MIN_MAX ?
  branch(this, start_order, Gecode::BVAR_MIN_MIN, Gecode::BVAL_MIN);
  branch(this, finish_order, Gecode::BVAR_MIN_MIN, Gecode::BVAL_MIN);
  // branch(this, start_product, Gecode::BVAR_MIN_MIN, Gecode::BVAL_MIN);
#endif
}

OpenStacksTime::OpenStacksTime(bool share, OpenStacksTime& ost)
  : Gecode::Space(share, ost),
    n_products(ost.n_products),
    n_orders(ost.n_orders),
    n_stacks(ost.n_stacks),
    t_max(ost.t_max),
    orders(ost.orders),
    duration(ost.duration)
{
  start_order.update(this, share, ost.start_order);
  finish_order.update(this, share, ost.finish_order);
  open_order.update(this, share, ost.open_order);
  start_product.update(this, share, ost.start_product);
#ifdef EXTRA_CARD_CONSTRAINTS
  finish_product.update(this, share, ost.finish_product);
  start_before_make.update(this, share, ost.start_before_make);
  finish_after_make.update(this, share, ost.finish_after_make);
  n_before.update(this, share, n_before);
  n_after.update(this, share, n_after);
#endif
}

Gecode::Space* OpenStacksTime::copy(bool share)
{
  return new OpenStacksTime(share, *this);
}

void OpenStacksTime::print()
{
  for (HSPS::index_type k = 0; k < n_orders; k++) {
    std::cerr << "start order #" << k + 1 << ": ["
	      << start_order[k].min() << ", " << start_order[k].max() << "]"
	      << std::endl;
    std::cerr << "finish order #" << k + 1 << ": ["
	      << start_order[k].min() + open_order[k].min()
	      << ", " << start_order[k].max() + open_order[k].max() << "]"
	      << std::endl;
  }
  for (HSPS::index_type k = 0; k < n_products; k++)
    std::cerr << "start make #" << k + 1 << ": ["
	      << start_product[k].min() << ", " << start_product[k].max()
	      << "]" << std::endl;
}

void OpenStacksTime::construct
(const graph& cdg,
 bool_vec& o_open,
 bool_vec& o_done)
{
//   if ((t_now + o_open.count(true)) >= (t_max - 1)) {
//     n_backtracks += 1;
//     return;
//   }
  bool finished = true;
  for (index_type k = 0; k < n_orders; k++) if (o_open[k] && !o_done[k]) {
    finished = false;
    bool can_end = true;
    for (index_type i = 0; (i < cdg.bi_degree(k)) && can_end; i++)
      if (!o_open[cdg.bidirectional(k)[i]])
	can_end = false;
    if (can_end) {
      int t_min = finish_order[k].min();
      //std::cerr << "trying finish[" << k + 1 << "] = " << t_min << std::endl;
      OpenStacksTime* new_s = (OpenStacksTime*)clone(false);
      eq(new_s, new_s->finish_order[k], t_min);
//       for (index_type i = 0; i < n_orders; i++)
//  	if ((i != k) && !o_open[i] && !o_done[i])
//  	  post(new_s, new_s->start_order[i] > t_min);
// 	else if ((i != k) && o_open[i] && !o_done[i])
//  	  post(new_s, new_s->finish_order[i] > t_min);
      if (new_s->status() != Gecode::SS_FAILED) {
	o_open[k] = false;
	o_done[k] = true;
	new_s->construct(cdg, o_open, o_done);
	o_open[k] = true;
	o_done[k] = false;
      }
      else {
	n_backtracks += 1;
#ifdef STOP_AT_FIRST_FAIL
	std::cerr << "first failure!" << std::endl;
	new_s->print();
	std::cerr << "parent state:" << std::endl;
	print();
	exit(0);
#endif
      }
      delete new_s;
    }
  }
  for (index_type k = 0; k < n_orders; k++) if (!o_open[k] && !o_done[k]) {
    finished = false;
    int t_min = start_order[k].min();
    //std::cerr << "trying start[" << k + 1 << "] = " << t_min << std::endl;
    OpenStacksTime* new_s = (OpenStacksTime*)clone(false);
    eq(new_s, new_s->start_order[k], t_min);
//     for (index_type i = 0; i < n_orders; i++)
//       if ((i != k) && !o_open[i] && !o_done[i])
// 	post(new_s, (new_s->start_order[i] > t_min));
//       else if ((i != k) && o_open[i] && !o_done[i])
// 	post(new_s, new_s->finish_order[i] > t_min);
    if (new_s->status() != Gecode::SS_FAILED) {
      o_open[k] = true;
      new_s->construct(cdg, o_open, o_done);
      o_open[k] = false;
    }
    else {
      n_backtracks += 1;
#ifdef STOP_AT_FIRST_FAIL
      std::cerr << "first failure!" << std::endl;
      new_s->print();
      exit(0);
#endif
    }
    delete new_s;
  }
  if (finished) {
    std::cerr << "solution found!" << std::endl;
    print();
    std::cerr << n_backtracks << " backtracks" << std::endl;
    exit(0);
  }
}

int main(int argc, char** argv)
{
  HSPS::index_type    n_products;
  HSPS::index_set_vec orders;
  HSPS::index_vec     durations;
  HSPS::index_type    n_stacks;
  HSPS::index_type    t_max;

  if (argc < 3) {
    std::cerr << "error: need two arguments (#stacks and Tmax)" << std::endl;
    exit(1);
  }

  n_stacks = atoi(argv[1]);
  t_max = atoi(argv[2]);

  std::cerr << "reading problem..." << std::endl;
  read_basic_problem(n_products, orders);
  read_durations(n_products, durations);

  std::cerr << "creating model..." << std::endl;
  OpenStacksTime* ost = 0;
  try {
    ost = new OpenStacksTime(n_products, orders, durations, n_stacks, t_max);

    if (ost->status() == Gecode::SS_FAILED) {
      std::cerr << "status == FAILED (over before it started!)" << std::endl;
      exit(0);
    }

#ifdef MY_SEARCH
    std::cerr << "constructing co-demand graph..." << std::endl;
    weighted_graph cdg;
    build_codemand_graph(orders, cdg);
    std::cerr << "searching..." << std::endl;
    bool_vec o_open(false, orders.length());
    bool_vec o_done(false, orders.length());
    ost->construct(cdg, o_open, o_done);
    std::cerr << "no solution found" << std::endl;
#else
    std::cerr << "creating engine..." << std::endl;
    Gecode::DFS<OpenStacksTime> e(ost);

    std::cerr << "searching..." << std::endl;
    OpenStacksTime* sol = e.next();
    if (sol == NULL) {
      std::cerr << "no solution found" << std::endl;
    }
    else {
      std::cerr << "found a solution (status = " << sol->status() << ")"
		<< std::endl;
      sol->print();
    }
    std::cerr << "propagation steps: "
	      << e.statistics().propagate << std::endl;
    std::cerr << "failed nodes:      " << e.statistics().fail << std::endl;
    std::cerr << "clones:            " << e.statistics().clone << std::endl;
    std::cerr << "commit operations: " << e.statistics().commit << std::endl;
    std::cerr << "peak memory:       " << e.statistics().memory << std::endl;
#endif
  }
  catch (Gecode::Exception e) {
    std::cerr << "exception " << e.what() << " caught" << std::endl;
    exit(1);
  }
}
