
#include "soft_solver.h"
#include "problem_io.h"

#define ADD_LOWER_BOUND

bool bb_print_plan = false;
bool bb_verbose = false;
count_type bb_lin_calls = 0;
count_type bb_calls_limit = count_type_max;

index_type exp_loss(index_type n_rem, index_type o_rem)
{
  index_type sum = 0;
  index_type v = 1;
  for (index_type k = 1; k <= o_rem; k++) {
    if (k > n_rem) sum += v;
    v = (2 * v);
  }
  return sum;
}

void bb_exp_loss
(const index_set_vec& o, // (original) orders
 const index_vec& s,     // schedule (position -> product#)
 index_type m,           // max stacks
 const index_vec& invs,  // inverse schedule (product# -> position)
 index_vec& rem,         // order remaining size
 index_vec& f,           // first
 index_vec& l,           // last
 index_vec& n,           // #open
 index_type acc,         // accumulated loss
 index_type& U,          // upper bound on loss
 index_set_vec& o_min)   // min loss solution (modified orders)
{
  // find first position (p) where #open exceeds max (m)
  index_type p = no_such_index;
  for (index_type k = 0; (k < s.length()) && (p == no_such_index); k++)
    if (n[k] > m) p = k;

  // if no such p, no need to drop any more deliveries
  if (p == no_such_index) {
    if (bb_verbose) {
      std::cerr << "solution with loss " << acc << ": " << rem << std::endl;
    }
    if ((acc < U) || (o_min.length() == 0)) {
      o_min.assign_value(EMPTYSET, o.length());
      for (index_type k = 0; k < o.length(); k++) {
	for (index_type i = 0; i < rem[k]; i++)
	  o_min[k].insert(o[k][i]);
      }
      if (bb_print_plan) {
	write_sp_plan(std::cout, o_min, s, m, false);
      }
    }
    if (acc < U) {
      U = acc;
    }
    return;
  }

  // else, examine options for reducing #open at p...
  for (index_type k = 0; k < o.length(); k++)
    if ((rem[k] > 0) && (f[k] <= p) && (l[k] >= p)) {
      for (index_type i = 0; i < rem[k]; i++) {
	index_type loss_i = exp_loss(i, rem[k]);
	if (acc + loss_i <= U) {
	  if (i == 0) {
	    index_type f_k = f[k];
	    index_type l_k = l[k];
	    index_type rem_k = rem[k];
	    f[k] = no_such_index;
	    l[k] = no_such_index;
	    rem[k] = 0;
	    index_vec new_n;
	    compute_open_profile(f, l, s.length(), new_n);
	    bb_exp_loss(o, s, m, invs, rem, f, l, new_n, acc + loss_i, U, o_min);
	    f[k] = f_k;
	    l[k] = l_k;
	    rem[k] = rem_k;
	  }
	  else {
	    index_type f_no_i = invs[o[k][0]]; // first pos of o[k][0..i-1]
	    index_type l_no_i = invs[o[k][0]]; // last pos of o[k][0..i-1]
	    for (index_type j = 1; j < i; j++) {
	      if (invs[o[k][j]] < f_no_i) f_no_i = invs[o[k][j]];
	      if (invs[o[k][j]] > l_no_i) l_no_i = invs[o[k][j]];
	    }
	    if ((f_no_i > p) || (l_no_i < p)) {
	      index_type f_k = f[k];
	      index_type l_k = l[k];
	      index_type rem_k = rem[k];
	      f[k] = f_no_i;
	      l[k] = l_no_i;
	      rem[k] = i;
	      index_vec new_n;
	      compute_open_profile(f, l, s.length(), new_n);
	      bb_exp_loss(o, s, m, invs, rem, f, l, new_n, acc + loss_i, U, o_min);
	      f[k] = f_k;
	      l[k] = l_k;
	      rem[k] = rem_k;
	    }
	  }
	}
      }
    }
}

index_type min_exp_loss
(const index_set_vec& o,
 const index_vec& s,
 index_type m,
 index_type U,
 index_set_vec& o_min)
{
  index_vec invs;
  mapping::invert_map(s, invs);
  assert(invs.length() == s.length());
  index_vec f;
  index_vec l;
  index_vec n;
  compute_open_schedule(o, s, f, l, n);
  index_vec rem(0, o.length());
  for (index_type k = 0; k < o.length(); k++)
    rem[k] = o[k].length();
  index_type mU = U;
  bb_exp_loss(o, s, m, invs, rem, f, l, n, 0, mU, o_min);
  return mU;
}

void bb_lin_loss_apply
(index_set_vec& o,       // orders (current; may be modified)
 const index_vec& s,     // schedule (position -> product#)
 index_type m,           // max stacks
 const index_vec& invs,  // inverse schedule (product# -> position)
 index_type acc,         // accumulated loss
 index_type& U,          // upper bound on loss
 index_set_vec& o_min,   // min loss solution (modified orders)
 index_vec& open_at_p,
 index_set_vec& left_of_p,
 index_set_vec& right_of_p,
 index_type n_to_apply,
 index_type next_may_apply)
{
  if (n_to_apply == 0) {
    index_vec new_f;
    index_vec new_l;
    index_vec new_n;
    compute_open_schedule(o, s, new_f, new_l, new_n);
    bb_lin_loss(o, s, m, invs, new_f, new_l, new_n, acc, U, o_min);
    return;
  }

  index_type n_rest = (n_to_apply - 1);

  bool_vec tried_left(false, open_at_p.length());
  bool_vec tried_right(false, open_at_p.length());

  while (true) {
    index_type k_min = no_such_index;
    bool k_min_is_left = true;
    index_type c_min = index_type_max;

    for (index_type k = next_may_apply; k < (open_at_p.length() - n_rest); k++) {
      if (!tried_left[k]) {
	index_type mc_rest = 0;
	if (n_rest > 0) {
	  index_vec c_fix(0, 0);
	  for (index_type i = k + 1; i < open_at_p.length(); i++) {
	    index_type mc_i = (left_of_p[i].length() < right_of_p[i].length() ?
			       left_of_p[i].length() : right_of_p[i].length());
	    c_fix.insert_ordered(mc_i, index_vec_util::increasing);
	  }
	  for (index_type i = 0; i < n_rest; i++)
	    mc_rest += c_fix[i];
	}
	if ((left_of_p[k].length() + mc_rest) < c_min) {
	  k_min = k;
	  k_min_is_left = true;
	  c_min = (left_of_p[k].length() + mc_rest);
	}
      }
      if (!tried_right[k]) {
	index_type mc_rest = 0;
	if (n_rest > 0) {
	  index_vec c_fix(0, 0);
	  for (index_type i = k + 1; i < open_at_p.length(); i++) {
	    index_type mc_i = (left_of_p[i].length() < right_of_p[i].length() ?
			       left_of_p[i].length() : right_of_p[i].length());
	    c_fix.insert_ordered(mc_i, index_vec_util::increasing);
	  }
	  for (index_type i = 0; i < n_rest; i++)
	    mc_rest += c_fix[i];
	}
	if ((right_of_p[k].length() + mc_rest) < c_min) {
	  k_min = k;
	  k_min_is_left = false;
	  c_min = (right_of_p[k].length() + mc_rest);
	}
      }
    }

    if (k_min == no_such_index) return;
    if ((acc + c_min) > U) return;
    if (bb_lin_calls > bb_calls_limit) return;

    if (k_min_is_left) {
      o[open_at_p[k_min]].subtract(left_of_p[k_min]);
      bb_lin_loss_apply(o, s, m, invs, acc + left_of_p[k_min].length(), U, o_min, open_at_p, left_of_p, right_of_p, n_to_apply - 1, k_min + 1);
      o[open_at_p[k_min]].insert(left_of_p[k_min]);
      tried_left[k_min] = true;
    }
    else {
      o[open_at_p[k_min]].subtract(right_of_p[k_min]);
      bb_lin_loss_apply(o, s, m, invs, acc + right_of_p[k_min].length(), U, o_min, open_at_p, left_of_p, right_of_p, n_to_apply - 1, k_min + 1);
      o[open_at_p[k_min]].insert(right_of_p[k_min]);
      tried_right[k_min] = true;
    }
  }
}

void bb_lin_loss
(index_set_vec& o,       // orders (current; may be modified)
 const index_vec& s,     // schedule (position -> product#)
 index_type m,           // max stacks
 const index_vec& invs,  // inverse schedule (product# -> position)
 index_vec& f,           // first
 index_vec& l,           // last
 index_vec& n,           // #open
 index_type acc,         // accumulated loss
 index_type& U,          // upper bound on loss
 index_set_vec& o_min)   // min loss solution (modified orders)
{
  // find all positions (ps) where #open exceeds max (m) and is maximal
  index_set ps;
  index_type mp = 0;
  for (index_type k = 0; k < s.length();k++) if (n[k] > m) {
    if (n[k] > mp) {
      mp = n[k];
      ps.assign_singleton(k);
    }
    else if (n[k] == mp) {
      ps.insert(k);
    }
  }

  // if no such p, no need to drop any more deliveries
  if (ps.empty()) {
    if ((acc < U) || (o_min.length() == 0)) {
      if (bb_verbose) {
	std::cerr << "solution with loss " << acc
		  << " (" << bb_lin_calls << " calls)"
		  << std::endl;
      }
      o_min.assign_copy(o);
      if (bb_print_plan) {
	write_sp_plan(std::cout, o_min, s, m, false);
      }
    }
    if (acc < U) {
      U = acc;
    }
    return;
  }

  assert(mp > m);
  index_type n_app = (mp - m);

  bb_lin_calls += 1;
  if (bb_lin_calls > bb_calls_limit) return;

  if ((acc + n_app) > U) {
    // std::cerr << "fast cut!" << std::endl;
    return;
  }

  // get some lower bounds on the cost of fixing each problematic p
  index_type p_max = no_such_index;
  index_type mc_max = 0;
  for (index_type p = 0; p < ps.length(); p++) {
#ifdef TRACE_PRINT_LOTS
    std::cerr << "p = " << ps[p] << std::endl;
#endif
    assert(n[ps[0]] == mp);
    index_vec c_fix(0, 0);
    for (index_type k = 0; k < o.length(); k++)
      if ((f[k] <= ps[p]) && (l[k] >= ps[p])) {
	index_type n_left_of_p = 0;
	index_type n_right_of_p = 0;
	for (index_type i = 0; i < o[k].length(); i++) {
	  if (invs[o[k][i]] <= ps[p]) n_left_of_p += 1;
	  if (invs[o[k][i]] >= ps[p]) n_right_of_p += 1;
	}
	if (n_left_of_p < n_right_of_p)
	  c_fix.insert_ordered(n_left_of_p, index_vec_util::increasing);
	else
	  c_fix.insert_ordered(n_right_of_p, index_vec_util::increasing);
      }
    assert(c_fix.length() > n_app);
    index_type mc_tot = 0;
    for (index_type k = 0; k < n_app; k++)
      mc_tot += c_fix[k];
#ifdef TRACE_PRINT_LOTS
    std::cerr << "c_fix = " << c_fix
	      << ", n_app = " << n_app
	      << ", mc_tot = " << mc_tot
	      << std::endl;
#endif
    if ((p_max == no_such_index) || (mc_tot > mc_max)) {
      p_max = p;
      mc_max = mc_tot;
    }
  }

  // let p be the most costly to fix problem position
  assert(p_max < ps.length());
  index_type p = ps[p_max];

  if ((acc + mc_max) > U) {
    // std::cerr << "cut!" << std::endl;
    return;
  }

#ifdef ADD_LOWER_BOUND
  // get some better lower bounds...
  index_type mc_max_add = 0;
  for (index_type q = 0; q < s.length(); q++) if ((n[q] > m) && (q != p)) {
    index_type no = 0; // # of overlapping open orders
    index_vec c_fix(0, 0);
    for (index_type k = 0; (k < o.length()) && ((n[q] - no) > m); k++)
      if ((f[k] <= q) && (l[k] >= q)) {
	if ((f[k] > p) || (l[k] < p)) {
	  index_type n_left_of_q = 0;
	  index_type n_right_of_q = 0;
	  for (index_type i = 0; i < o[k].length(); i++) {
	    if (invs[o[k][i]] <= q) n_left_of_q += 1;
	    if (invs[o[k][i]] >= q) n_right_of_q += 1;
	  }
	  if (n_left_of_q < n_right_of_q)
	    c_fix.insert_ordered(n_left_of_q, index_vec_util::increasing);
	  else
	    c_fix.insert_ordered(n_right_of_q, index_vec_util::increasing);
	}
	else {
	  no += 1;
	}
      }
    assert(c_fix.length() >= ((n[q] - no) - m));
    index_type mc_add = 0;
    for (index_type k = 0; k < ((n[q] - no) - m); k++)
      mc_add += c_fix[k];
    if (mc_add > mc_max_add) mc_max_add = mc_add;
  }

  if ((acc + mc_max + mc_max_add) > U) {
    // std::cerr << "cut!" << std::endl;
    return;
  }
#endif

  // now, examine options for reducing #open at p...

  index_vec open_at_p(no_such_index, 0);
  index_set_vec left_of_p(EMPTYSET, 0);
  index_set_vec right_of_p(EMPTYSET, 0);

  for (index_type k = 0; k < o.length(); k++) if ((f[k] <= p) && (l[k] >= p)) {
    open_at_p.append(k);
    index_type j = (open_at_p.length() - 1);
    left_of_p.append(EMPTYSET);
    assert(left_of_p.length() == open_at_p.length());
    right_of_p.append(EMPTYSET);
    assert(right_of_p.length() == open_at_p.length());
    for (index_type i = 0; i < o[k].length(); i++) {
      if (invs[o[k][i]] <= p) left_of_p[j].insert(o[k][i]);
      if (invs[o[k][i]] >= p) right_of_p[j].insert(o[k][i]);
    }
    assert(!left_of_p[j].empty());
    assert(!right_of_p[j].empty());
  }

  bb_lin_loss_apply(o, s, m, invs, acc, U, o_min, open_at_p, left_of_p, right_of_p, n_app, 0);
  // bb_lin_loss_apply(o, s, m, invs, acc, U, o_min, open_at_p, left_of_p, right_of_p, 1, 0);
}

index_type min_lin_loss
(const index_set_vec& o,
 const index_vec& s,
 index_type m,
 index_type U,
 index_set_vec& o_min)
{
  index_vec invs;
  mapping::invert_map(s, invs);
  assert(invs.length() == s.length());
  index_vec f;
  index_vec l;
  index_vec n;
  compute_open_schedule(o, s, f, l, n);
  index_set_vec mo(o);
  index_type mU = U;
  bb_lin_calls = 0;
  bb_lin_loss(mo, s, m, invs, f, l, n, 0, mU, o_min);
  if (bb_verbose) {
    std::cerr << "final upper bound = " << mU
	      << " (" << bb_lin_calls << " calls)"
	      << std::endl;
  }
  return mU;
}
