
#include "basic_solver.h"
#include "soft_solver.h"
#include "problem_io.h"
#include "stats.h"

// print stats summary every 10000 solutions
// #define RUNNING_TRACE_PRINT

int main(int argc, char* argv[])
{
  index_type    original_n_products;
  index_set_vec original_orders;
  index_type    reduced_n_products;
  index_set_vec reduced_orders;
  index_set_vec reduced_product_map;
  index_vec     duration;
  // index_matrix  costs;
  index_type    lb = 0;
  index_type    ub = 0;
  bool          opt_try = false;
  bool          opt_all = false;
  bool          opt_incremental = false;
  bool          opt_preprocess = false;
  bool          opt_plan = false;
  bool          opt_inv_count = false;
  bool          opt_print_schedule = false;
  bool          opt_peak_depth = false;
  bool          opt_soft_exp = false;
  bool          opt_soft_lin = false;
  bool          opt_makespan = false;
  index_type    soft_stack_limit = 0;
  index_type    soft_max_loss = 0;
  bool          opt_verbose = false;
  double        time_limit = 0;

  for (index_type k = 0; k < argc; k++) {
    if ((strcmp(argv[k],"-l") == 0) && (k < (argc - 1))) {
      lb = atoi(argv[++k]);
    }
    else if ((strcmp(argv[k],"-u") == 0) && (k < (argc - 1))) {
      ub = atoi(argv[++k]);
    }
    else if (strcmp(argv[k],"-try") == 0) {
      opt_try = true;
    }
    else if (strcmp(argv[k],"-all") == 0) {
      opt_all = true;
    }
    else if (strcmp(argv[k],"-inc") == 0) {
      opt_incremental = true;
    }
    else if (strcmp(argv[k],"-prep") == 0) {
      opt_preprocess = true;
    }
    else if (strcmp(argv[k],"-s") == 0) {
      opt_print_schedule = true;
    }
    else if (strcmp(argv[k],"-peak") == 0) {
      opt_peak_depth = true;
    }
    else if (strcmp(argv[k],"-m") == 0) {
      opt_makespan = true;
    }
    else if (strcmp(argv[k],"-p") == 0) {
      opt_plan = true;
    }
    else if (strcmp(argv[k],"-ic") == 0) {
      opt_inv_count = true;
    }
    else if ((strcmp(argv[k],"-soft-exp") == 0) && (k < (argc - 2))) {
      opt_soft_exp = true;
      soft_stack_limit = atoi(argv[++k]);
      soft_max_loss = atoi(argv[++k]);
    }
    else if ((strcmp(argv[k],"-soft-lin") == 0) && (k < (argc - 2))) {
      opt_soft_lin = true;
      soft_stack_limit = atoi(argv[++k]);
      soft_max_loss = atoi(argv[++k]);
    }
    else if ((strcmp(argv[k],"-c") == 0) && (k < (argc - 1))) {
      bb_calls_limit = atoi(argv[++k]);
    }
    else if (strcmp(argv[k],"-v") == 0) {
      opt_verbose = true;
    }
    else if ((strcmp(argv[k],"-t") == 0) && (k < argc - 1)) {
      time_limit = atof(argv[++k]);
    }
  }

  Stopwatch stats;

  read_basic_problem(original_n_products, original_orders);
  if (opt_makespan)
    read_durations(original_n_products, duration);

  std::cerr << "problem read" << std::endl;
  std::cerr << "#products = " << original_n_products << std::endl;
  std::cerr << "#orders = " << original_orders.length() << std::endl;
  for (index_type k = 0; k < original_orders.length(); k++) {
    std::cerr << "order #" << k << ": " << original_orders[k] << std::endl;
  }
  if (opt_makespan) {
    std::cerr << "durations = " << duration << std::endl;
  }
  if (opt_preprocess) {
    preprocess(original_n_products, original_orders,
	       reduced_n_products, reduced_orders,
	       reduced_product_map);
  }
  else {
    reduced_n_products = original_n_products;
    reduced_orders.assign_copy(original_orders);
  }

  if (lb == 0) lb = 1;
  if (ub == 0) ub = reduced_orders.length();

  BasicSolver* solver = new BasicSolver(reduced_n_products, reduced_orders);
  index_type min_value = no_such_index;
  index_vec s_ext;
  index_vec schedule;
  bool ok = false;
  index_type soft_min_loss = soft_max_loss + 1;
  NTYPE min_makespan = POS_INF;

  if (opt_try) {
    ub = reduced_orders.length();
    bool opt = false;
    stats.start();
    while (!opt) {
      std::cerr << "trying U = " << ub << "..." << std::endl;
      solver->clear();
      index_type v = solver->stacks(ub, ub, false);
      std::cerr << "min = " << v << " (" << stats.time() << " seconds, "
		<< stats.peak_memory() << "k)"
		<< std::endl;
      if (v > ub) {
	opt = true;
      }
      else {
	min_value = v;
	ok = solver->extract(min_value, s_ext);
	if (!ok) {
	  std::cerr << "error: failed to extract schedule!" << std::endl;
	  exit(255);
	}
	ub = v - 1;
      }
    }
    stats.stop();
  }
  else if ((lb < ub) || opt_incremental) {
    std::cerr << "running solver..." << std::endl;
    stats.start();
    min_value = solver->stacks(lb, ub, false);
    ok = solver->extract(min_value, s_ext);
    if (!ok) {
      std::cerr << "error: failed to extract schedule!" << std::endl;
      exit(255);
    }
    stats.stop();
    std::cerr << "min = " << min_value << " (" << stats.time() << " seconds, "
	      << stats.peak_memory() << "k)"
	      << std::endl;
  }
  else {
    min_value = ub;
  }

  std::cerr << "min max #open stacks: " << min_value << std::endl;

  if (!opt_incremental) {
    if (opt_all) {
      std::cerr << "searching for all solutions..." << std::endl;
      stats.start();
      solver->clear();
      solver->stacks(min_value, min_value, true);
      ok = solver->extract(min_value, s_ext);
      stats.stop();
    }
    if (ok) {
      std::cerr << "first solution found (" << stats.time() << " seconds)"
		<< std::endl;
    }
    else {
      std::cerr << "no solution found (" << stats.time() << " seconds)"
		<< std::endl;
    }
  }

  stats.start();
  index_type n = 0;
  index_type n_last_print = 0;
  count_type total_bb_calls = 0;
  index_type min_peak_depth = index_type_max;
  index_type max_peak_depth = 0;
  while (ok) {
    if (opt_preprocess) {
      postprocess(s_ext, reduced_product_map, original_n_products, schedule);
    }
    else {
      schedule.assign_copy(s_ext);
    }
    if (!opt_all || opt_verbose ||
	(!opt_soft_exp && !opt_soft_lin && !opt_makespan &&
	 !opt_plan && !opt_print_schedule && !opt_peak_depth)) {
      std::cerr << "product schedule #" << n << ": " << schedule
		<< std::endl;
    }
    if (opt_print_schedule || opt_peak_depth) {
      index_vec f;
      index_vec l;
      index_vec n;
      compute_open_schedule(original_orders, schedule, f, l, n);
      if (opt_print_schedule) {
	std::cout << "product schedule #" << n << ": " << schedule
		  << std::endl;
	index_vec invs;
	mapping::invert_map(schedule, invs);
	std::cout << "inv. schedule: " << invs << std::endl;
	std::cout << "f = " << f << std::endl;
	std::cout << "l = " << l << std::endl;
	std::cout << "n = " << n << std::endl;
	write_open_schedule(std::cout, original_orders, schedule, 0);
      }
      if (opt_peak_depth) {
	index_type d = n.first(min_value);
	assert(d != no_such_index);
	if (opt_verbose) {
	  std::cout << "peak depth = " << d + 1 << std::endl;
	}
	if (d < min_peak_depth)
	  min_peak_depth = d;
	if (d > max_peak_depth)
	  max_peak_depth = d;
      }
    }

    if (opt_makespan) {
      weighted_graph g((2 * original_orders.length()) + original_n_products);
      #define START(k) k
      #define FINISH(k) (original_orders.length() + k)
      #define MAKE(k) ((2 * original_orders.length()) + k)
      for (index_type k = 0; k < original_orders.length(); k++)
	g.set_weight(START(k), 1);
      for (index_type k = 0; k < original_orders.length(); k++)
	g.set_weight(FINISH(k), 1);
      for (index_type k = 0; k < original_n_products; k++)
	g.set_weight(MAKE(k), duration[k]);
      for (index_type k = 0; k < original_orders.length(); k++)
	for (index_type i = 0; i < original_orders[k].length(); i++) {
	  g.add_edge(START(k), MAKE(original_orders[k][i]), 0);
	  g.add_edge(MAKE(original_orders[k][i]), FINISH(k), 0);
	}
      index_vec f;
      index_vec l;
      index_vec n;
      compute_open_schedule(original_orders, schedule, f, l, n);
      for (index_type i = 0; i < original_orders.length(); i++)
	for (index_type j = 0; j < original_orders.length(); j++)
	  if (l[i] < f[j])
	    g.add_edge(FINISH(i), START(j), 0);
      cost_vec start;
      NTYPE m = g.critical_path(start);
      if ((m < min_makespan) || opt_verbose) {
	std::cerr << "start times = " << start << std::endl;
	std::cerr << "makespan = " << m << std::endl;
	if (m < min_makespan) {
	  min_makespan = m;
	}
      }
    }

    else if (opt_soft_exp) {
      if (!opt_all || opt_verbose) {
	std::cerr << "computing min loss (exp.) with stack limit = "
		  << soft_stack_limit << " and bound = "
		  << (soft_min_loss < soft_max_loss ?
		      soft_min_loss : soft_max_loss)
		  << "..." << std::endl;
      }
      index_set_vec o_min(EMPTYSET, 0);
      index_type v = min_exp_loss(original_orders, schedule, soft_stack_limit,
				  (soft_min_loss < soft_max_loss ?
				   soft_min_loss : soft_max_loss), o_min);
      if (o_min.length() == 0) {
	if (!opt_all || opt_verbose) {
	  std::cerr << "no solution within this stack limit and bound!"
		    << std::endl;
	}
      }
      else if ((v < soft_min_loss)  || opt_verbose) {
	std::cerr << "min loss = " << v << std::endl;
	soft_min_loss = v;
	assert(o_min.length() == original_orders.length());
	if (opt_plan) {
	  write_sp_plan(std::cout, o_min, schedule,
			soft_stack_limit, opt_inv_count);
	}
      }
    }

    else if (opt_soft_lin) {
      if (!opt_all || opt_verbose) {
	std::cerr << "computing min loss (linear) with stack limit = "
		  << soft_stack_limit << " and bound = "
		  << (soft_min_loss < soft_max_loss ?
		      soft_min_loss : soft_max_loss)
		  << "..." << std::endl;
      }
      index_set_vec o_min(EMPTYSET, 0);
      index_type v = min_lin_loss(original_orders, schedule, soft_stack_limit,
				  (soft_min_loss < soft_max_loss ?
				   soft_min_loss : soft_max_loss), o_min);
      total_bb_calls += bb_lin_calls;
      if (o_min.length() == 0) {
	if (!opt_all || opt_verbose) {
	  std::cerr << "no solution within this stack limit and bound! ("
		    << bb_lin_calls << " calls, " << stats.time() << " seconds)"
		    << std::endl;
	}
      }
      else if ((v < soft_min_loss) || opt_verbose) {
	std::cerr << "min loss = " << v
		  << " (" << bb_lin_calls << " calls, "
		  << stats.time() << " seconds)"
		  << std::endl;
	soft_min_loss = v;
	assert(o_min.length() == original_orders.length());
	if (opt_plan) {
	  write_sp_plan(std::cout, o_min, schedule,
			soft_stack_limit, opt_inv_count);
	}
      }
      else if (opt_verbose) {
	std::cerr << "min loss = " << v
		  << " (" << bb_lin_calls << " calls, "
		  << stats.time() << " seconds)"
		  << std::endl;
      }
    }

    else if (opt_plan) {
      write_seq_plan(std::cout, original_orders, schedule, min_value);
    }

    if (opt_all) {
      if (time_limit > 0) {
	if (stats.total_time() > time_limit) {
	  std::cerr << "time-out!" << std::endl;
	  exit(0);
	}
      }
      // index_vec s_check(s_ext);
      ok = solver->extract_next(min_value, s_ext);
      if (!ok && opt_incremental) {
	// assert(s_ext == s_check);
	std::cerr << "searching for more solutions..." << std::endl;
	solver->morestacks(min_value, min_value, 10);
	ok = solver->extract_next(min_value, s_ext);
      }
      n += 1;
#ifdef RUNNING_TRACE_PRINT
      if (n >= (n_last_print + 10000)) {
	std::cerr << "n = " << n << " (" << (total_bb_calls/((double)n))
		  << " calls avg., " << stats.time() << " seconds)"
		  << std::endl;
	n_last_print = n;
      }
#endif
    }
    else {
      ok = false;
    }
  }
  stats.stop();

  std::cerr << n << " solutions found (" << stats.total_time() << " seconds)"
	    << std::endl;
  std::cerr << "peak depth: min = " << min_peak_depth
	    << ", max = " << max_peak_depth << std::endl;
}
