
#include "basic_solver.h"
#include "soft_solver.h"
#include "problem_io.h"
#include "stats.h"

#include <fstream>

#define VERBOSE

void serialise2
(const weighted_graph& cdg,
 const index_vec&      d_min,
 index_type            d_max,
 index_set_vec&        ser,
 index_vec&            s_min,
 bool_vec&             solo,
 index_type            f1,
 index_type            f2,
 index_type            n_par,
 index_type            n_max,
 index_type&           best,
 count_type&           n_bt)
{
  index_type min1 = index_vec_util::max(s_min, 0);
  index_type min = (d_max > min1 ? d_max : min1);
  if (n_par <= n_max) {
    if (min < best) {
#ifdef VERBOSE
      std::cerr << "new solution found (2): " << min
		<< " (" << n_bt << " backtracks)" << std::endl;
      std::cerr << "ser = " << ser << std::endl;
#endif
      best = min;
    }
    return;
  }
  // add currently solo order to serialised set
  for (index_type i = f1; i < cdg.size(); i++) if (solo[i])
    for (index_type j = f2; j < ser.length(); j++)
      if (!cdg.adjacent(i, ser[j]) && ((d_min[i] + s_min[j]) < best)) {
	ser[j].insert(i);
	index_type r_j = s_min[j];
	s_min[j] = (s_min[j] + d_min[i]);
	solo[i] = false;
	// std::cerr << "2/" << i << "/" << j << ": serialising ("
	//	  << ser << ", " << s_min << ")" << std::endl;
	serialise2(cdg, d_min, d_max, ser, s_min, solo, i, j, n_par - 1, n_max, best, n_bt);
	assert(min <= best);
	ser[j].subtract(i);
	s_min[j] = r_j;
	solo[i] = true;
	// std::cerr << "2/" << i << "/" << j << ": retracting" << std::endl;
      }
  n_bt += 1;
}

void serialise1
(const weighted_graph& cdg,
 const index_vec&      d_min,
 index_type            d_max,
 index_set_vec&        ser,
 index_vec&            s_min,
 bool_vec&             solo,
 index_type            f1,
 index_type            n_par,
 index_type            n_max,
 index_type&           best,
 count_type&           n_bt)
{
  index_type min1 = index_vec_util::max(s_min, 0);
  index_type min = (d_max > min1 ? d_max : min1);
  if (n_par <= n_max) {
    if (min < best) {
#ifdef VERBOSE
      std::cerr << "new solution found (1): " << min
		<< " (" << n_bt << " backtracks)" << std::endl;
      std::cerr << "ser = " << ser << std::endl;
#endif
      best = min;
    }
    return;
  }
  // serialise two currently solo orders
  for (index_type i = f1; i < cdg.size(); i++) if (solo[i])
    for (index_type j = i + 1; j < cdg.size(); j++) if (solo[j])
      if (!cdg.adjacent(i, j) && ((d_min[i] + d_min[j]) < best)) {
	index_set& ns = ser.append();
	ns.insert(i);
	ns.insert(j);
	s_min.append(d_min[i] + d_min[j]);
	solo[i] = false;
	solo[j] = false;
	// std::cerr << "1/" << i << "/" << j << ": serialising ("
	//	  << ser << ", " << s_min << ")" << std::endl;
	serialise1(cdg, d_min, d_max, ser, s_min, solo, i, n_par - 1, n_max, best, n_bt);
	assert(min <= best);
	s_min.dec_length();
	ser.dec_length();
	solo[i] = true;
	solo[j] = true;
	// std::cerr << "1/" << i << "/" << j << ": retracting ("
	//	  << ser << ", " << s_min << ")" << std::endl;
      }
  serialise2(cdg, d_min, d_max, ser, s_min, solo, 0, 0, n_par, n_max, best, n_bt);
}

int main(int argc, char* argv[])
{
  index_type    n_products;
  index_set_vec orders;
  index_matrix  costs;
  index_vec     duration;
  index_type    lb = 0;
  index_type    ub = 0;
  bool          opt_preprocess = false;
  bool          opt_approximate = false;
  bool          opt_write_graphs = false;
  bool          opt_soft_lin = false;
  bool          opt_soft_exp = false;
  bool          opt_makespan = false;
  index_type    max_stack_limit = 0;
  index_type    min_stack_limit = 0;
  bool          opt_qp = false;
  bool          opt_mt = false;
  bool          opt_plan = false;
  bool          opt_inv = false;
  NTYPE         price_per_stack = 0;

  for (index_type k = 0; k < argc; k++) {
    if ((strcmp(argv[k],"-l") == 0) && (k < (argc - 1))) {
      lb = atoi(argv[++k]);
    }
    else if ((strcmp(argv[k],"-u") == 0) && (k < (argc - 1))) {
      ub = atoi(argv[++k]);
    }
    else if (strcmp(argv[k],"-prep") == 0) {
      opt_preprocess = true;
    }
    else if (strcmp(argv[k],"-apx") == 0) {
      opt_approximate = true;
    }
    else if (strcmp(argv[k],"-g") == 0) {
      opt_write_graphs = true;
    }
    else if ((strcmp(argv[k],"-soft-lin") == 0) && (k < (argc - 1))) {
      opt_soft_lin = true;
      max_stack_limit = atoi(argv[++k]);
      min_stack_limit = max_stack_limit;
    }
    else if ((strcmp(argv[k],"-soft-exp") == 0) && (k < (argc - 1))) {
      opt_soft_exp = true;
      max_stack_limit = atoi(argv[++k]);
      min_stack_limit = max_stack_limit;
    }
    else if ((strcmp(argv[k],"-qp") == 0) && (k < (argc - 1))) {
      opt_qp = true;
      min_stack_limit = 1;
      price_per_stack = A_TO_N(argv[++k]);
    }
    else if ((strcmp(argv[k],"-makespan") == 0) && (k < (argc - 1))) {
      opt_makespan = true;
      max_stack_limit = atoi(argv[++k]);
      min_stack_limit = max_stack_limit;
    }
    else if ((strcmp(argv[k],"-mt") == 0) && (k < (argc - 2))) {
      opt_mt = true;
      opt_makespan = true;
      min_stack_limit = atoi(argv[++k]);
      price_per_stack = A_TO_N(argv[++k]);
    }
    else if (strcmp(argv[k],"-p") == 0) {
      opt_plan = true;
    }
    else if (strcmp(argv[k],"-ic") == 0) {
      opt_inv = true;
    }
  }

  Stopwatch stats;

  read_weighted_problem(n_products, orders, costs);
  if (opt_makespan)
    read_durations(n_products, duration);
  if (opt_mt)
    max_stack_limit = orders.length();

  if (lb == 0) lb = 1;
  if (ub == 0) ub = orders.length();

  std::cerr << "problem read" << std::endl;
  std::cerr << "#products = " << n_products << std::endl;
  std::cerr << "#orders = " << orders.length() << std::endl;
  for (index_type k = 0; k < orders.length(); k++) {
    std::cerr << "order #" << k << ": " << orders[k] << std::endl;
  }
  // std::cerr << "request/cost matrix = " << costs << std::endl;

  weighted_graph cdg;
  build_weighted_codemand_graph(orders, costs, opt_soft_exp, cdg);

  // std::cerr << "co-demand graph: " << cdg << std::endl;
  // cdg.write_digraph(std::cerr, true, false, true, true, "Co-Demand Graph");

  if (opt_write_graphs) {
    std::cerr << "writing co-demand graph..." << std::endl;
    std::ofstream g_out("co_demand_graph.dot");
    write_codemand_graph(g_out, orders);
    g_out.close();
  }

  // do some stuff with min degree
  std::cerr << "min degree + 1 = " << cdg.min_bi_degree() + 1 << std::endl;
  std::cerr << "orders with min degree:";
  for (index_type k = 0; k < cdg.size(); k++)
    if (cdg.bi_degree(k) == cdg.min_bi_degree()) {
      std::cerr << " " << k << cdg.bidirectional(k);
    }
  std::cerr << std::endl;

  // do some stuff with cliques
  if (opt_approximate) {
    std::cerr << "finding cliques (approximate)..." << std::endl;
    graph c_cdg(cdg);
    c_cdg.complement();
    index_set v;
    index_type m = 0;
    for (index_type k = 0; k < orders.length(); k++) if (!v.contains(k)) {
      index_set c;
      c_cdg.apx_independent_set_including(k, c);
      assert(cdg.is_clique(c));
      // std::cerr << "size " << c.length() << " clique found: " << c << std::endl;
      if (c.length() > m) m = c.length();
      v.insert(c);
    }
    std::cerr << "max clique size = " << m << std::endl;
  }
  else {
    std::cerr << "finding cliques (optimal)..." << std::endl;
    index_set c;
    cdg.maximal_clique(c);
    assert(c.length() > 0);
    std::cerr << "max clique size = " << c.length() << std::endl;
  }

  // min loss analysis
  if (opt_soft_lin || opt_soft_exp) {
    NTYPE qp_min = POS_INF;

    NTYPE sc_min = POS_INF;
    for (index_type i = 0; i < orders.length(); i++)
      for (index_type j = 0; j < n_products; j++)
	if (costs[i][j] > 0) sc_min = MIN(sc_min, costs[i][j]);
    std::cerr << "smallest cost lower bound on loss: " << sc_min << std::endl;

    for (index_type lim = max_stack_limit; lim >= min_stack_limit; lim--) {
      std::cerr << "at stack limit " << lim << "..." << std::endl;
      NTYPE lb = sc_min;

      if ((cdg.min_bi_degree() + 1) > lim) {
	NTYPE c_min = 0;
	index_type n_over = 0;
	for (index_type k = 0; k < orders.length(); k++)
	  if (cdg.bi_degree(k) + 1 > lim) {
	    // std::cerr << "node " << k << " has "
	    //	  << ((cdg.bi_degree(k) + 1) - lim)
	    //	  << " too many neighbours" << std::endl;
	    cost_vec ec;
	    for (index_type i = 0; i < cdg.bidirectional(k).length(); i++)
	      ec.insert_ordered(cdg.edge_label(k, cdg.bidirectional(k)[i]),
				cost_vec_util::increasing);
	    // std::cerr << "sorted edge costs: " << ec << std::endl;
	    NTYPE c = 0;
	    for (index_type i = 0; i < ((cdg.bi_degree(k) + 1) - lim); i++)
	      c += ec[i];
	    // std::cerr << "cost of fixing node " << k << ": " << c
	    // << std::endl;
	    if ((c < c_min) || (n_over == 0)) c_min = c;
	    n_over += 1;
	  }
	std::cerr << "min degree lower bound on loss: " << c_min << std::endl;
	lb = MAX(lb, c_min);
      }

      if (opt_soft_lin) {
	index_set_vec cl;
	cdg.all_cliques_geq(lim + 1, cl);
	// std::cerr << "found " << cl.length() << " too large cliques"
	//	  << std::endl;
	if (cl.length() > 1) {
	  weighted_graph clig(cl.length());
	  for (index_type i = 0; i < cl.length(); i++) {
	    clig.node_label(i) = (cl[i].length() - lim);
	    for (index_type j = i + 1; j < cl.length(); j++) {
	      index_type n = cl[i].count_common(cl[j]);
	      if (n > 1)
		clig.add_undirected_edge(i, j, n);
	    }
	  }
	  clig.strongly_connected_components();
	  // std::cerr << clig.n_components()
	  //	    << " components in clique intersection graph"
	  //	    << std::endl;
	  NTYPE sum = 0;
	  for (index_type k = 0; k < clig.n_components(); k++) {
	    index_set cc;
	    clig.component_node_set(k, cc);
	    assert(cc.length() > 0);
	    cost_vec credit(0, cc.length());
	    NTYPE c_sum = 0;
	    for (index_type i = 0; i < cc.length(); i++) {
	      NTYPE l = (cl[cc[i]].length() - lim) - credit[i];
	      if (l > 0) {
		c_sum += l;
		for (index_type j = i + 1; j < cc.length(); j++)
		  if (clig.adjacent(i, j))
		    credit[j] += MIN(l, clig.edge_label(i, j));
	      }
	    }
	    // std::cerr << "min loss in component " << k << ": " << c_sum
	    //	      << std::endl;
	    sum += c_sum;
	  }
	  std::cerr << "clique lower bound on loss: " << sum << std::endl;
	  lb = MAX(lb, sum);
	}
	else if (cl.length() > 0) {
	  std::cerr << "clique lower bound on loss: " << cl[0].length() - lim
		    << std::endl;
	  lb = MAX(lb, cl[0].length() - lim);
	}
      }

      std::cerr << "max lower bound on loss at stack limit " << lim
		<< ": " << lb << std::endl;
      if (opt_qp) {
	std::cerr << "QP-bound at " << lim << ": "
		  << PRINT_NTYPE((lim * price_per_stack) + lb)
		  << std::endl;
	qp_min = MIN(qp_min, (lim * price_per_stack) + lb);
      }
    }

    if (opt_qp) {
      std::cerr << "lowest QP-bound: " << PRINT_NTYPE(qp_min) << std::endl;
    }
  }

  // do single-stack WIS construction if soft problem
  if (opt_soft_lin || opt_soft_exp) {
    index_set s1;
    NTYPE gain1 = cdg.apx_weighted_independent_set(s1);
    NTYPE loss1 = 0;
    for (index_type k = 0; k < orders.length(); k++)
      if (!s1.contains(k))
	loss1 += cdg.weight(k);
    std::cerr << "greedy single-stack construction (1): s = " << s1
	      << ", gain = " << gain1 << ", loss = " << loss1
	      << std::endl;
    // build "reduced" cdg: nodes in s1 have zero weight, nodes not in
    // s1 have their weight reduced by the cost of removing all edges to
    // nodes that are in s1; the value of an independent set in the rcdg
    // can be added to the first greedy solution
    weighted_graph rcdg(cdg);
    for (index_type k = 0; k < orders.length(); k++) {
      if (s1.contains(k)) {
	rcdg.set_weight(k, 0);
      }
      else {
	NTYPE wloss = 0;
	for (index_type i = 0; i < cdg.bidirectional(k).length(); i++)
	  if (s1.contains(cdg.bidirectional(k)[i])) {
	    if (opt_soft_exp)
	      wloss = MAX(wloss, cdg.weight(k, cdg.bidirectional(k)[i]));
	    else
	      wloss += cdg.weight(k, cdg.bidirectional(k)[i]);
	    rcdg.remove_undirected_edge(k, cdg.bidirectional(k)[i]);
	  }
	rcdg.set_weight(k, cdg.weight(k) - wloss);
      }
    }
    // std::cerr << "reduced cdg: " << rcdg << std::endl;
    index_set s2;
    NTYPE gain2 = rcdg.apx_weighted_independent_set(s2);
    assert(s2.first_common_element(s1) == no_such_index);
    std::cerr << "greedy single-stack construction (2): "
	      << ", gain = " << (gain1 + gain2)
	      << ", loss = " << (loss1 - gain2)
	      << std::endl;
    if (opt_plan) {
      // construct actual solution
      index_set_vec so(orders);
      for (index_type k = 0; k < orders.length(); k++)
	if (!s1.contains(k) && !s2.contains(k))
	  so[k].clear();
      for (index_type i = 0; i < s1.length(); i++)
	for (index_type j = 0; j < s2.length(); j++)
	  remove_common_requests(s1[i], so[s1[i]], s2[j], so[s2[j]], costs);
#ifdef VERBOSE
      std::cerr << "selected reduced orders:" << std::endl;
      for (index_type i = 0; i < s1.length(); i++)
	std::cerr << s1[i] << ": " << so[s1[i]] << std::endl;
      std::cerr << " -- " << std::endl;
      for (index_type i = 0; i < s2.length(); i++)
	std::cerr << s2[i] << ": " << so[s2[i]] << std::endl;
#endif
      index_vec s;
      for (index_type k = 0; k < so.length(); k++)
	for (index_type i = 0; i < so[k].length(); i++) {
	  assert(s.first(so[k][i]) == no_such_index);
	  s.append(so[k][i]);
	}
      assert(s.length() <= n_products);
      write_sp_plan(std::cout, so, s, 1, true);
    }
  }

  if (opt_makespan) {
    NTYPE mt_min = POS_INF;
    count_type n_bt = 0;

    stats.start();
    index_vec d_min(0, orders.length());
    for (index_type k = 0; k < orders.length(); k++)
      for (index_type i = 0; i < orders[k].length(); i++)
	if (duration[orders[k][i]] > d_min[k])
	  d_min[k] = duration[orders[k][i]];

    for (index_type lim = max_stack_limit; lim >= min_stack_limit; lim--) {
      std::cerr << "at stack limit " << lim << "..." << std::endl;

      index_set_vec ser;
      index_vec     s_min;
      bool_vec      solo(true, orders.length());
      index_type    best = index_type_max;

      serialise1(cdg, d_min, index_vec_util::max(d_min), ser, s_min, solo, 0, orders.length(), lim, best, n_bt);
      std::cerr << "lower bound on makespan: " << best
		<< " (" << n_bt << " backtracks, " << stats.time()
		<< " seconds)"
		<< std::endl;

      if (opt_mt) {
	std::cerr << "MT-bound at " << lim << ": "
		  << PRINT_NTYPE((lim * price_per_stack) + best)
		  << std::endl;
	mt_min = MIN(mt_min, (lim * price_per_stack) + best);
      }
    }

    if (opt_mt) {
      std::cerr << "lowest MT-bound: " << PRINT_NTYPE(mt_min) << std::endl;
    }
    stats.stop();
    std::cerr << n_bt << " backtracks in " << stats.time() << " seconds"
	      << std::endl;
  }
}
